// hbg2.cpp

#include <cstdlib>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include "optimize.h"
#include "kummer.h"

#define POL_DEG 4
#define POL_DEG_INC (POL_DEG+1)
double pol_coeff[POL_DEG_INC*POL_DEG_INC*POL_DEG_INC];
double delta_coeff[4][POL_DEG_INC*POL_DEG_INC*POL_DEG_INC];
double S_coeff[4][POL_DEG_INC*POL_DEG_INC*POL_DEG_INC];

using namespace std;
using namespace GiNaC;
using namespace boost::numeric;

extern symbol k1, k2, k3, k4;

template<typename T,typename P>
ostream& operator<<( ostream& os, const interval<T,P>& itv )
{
	return os << "[" << itv.lower() <<
		"," << itv.upper() << "]";
}


void ex_mpz(double *p_coeff, ex pol, int di)
{
	int i, j, k;
	exmap ms;
	ex x1, x2, x3, c;

	switch(di) {
		case 1:
			ms[k1] = 1;
			x1 = k2; x2 = k3; x3 = k4;
			break;
		case 2:
			ms[k2] = 1;
			x1 = k1; x2 = k3; x3 = k4;
			break;
		case 3:
			ms[k3] = 1;
			x1 = k1; x2 = k2; x3 = k4;
			break;
		case 4:
			ms[k4] = 1;
			x1 = k1; x2 = k2; x3 = k3;
			break;
		default: cerr << "error in ex_mpz" << endl; exit(1); break;
	}
	
	ex poldi = pol.subs(ms, subs_options::no_pattern);
		
	for(i = 0; i <= POL_DEG; i++) {
		for(j = 0; j <= POL_DEG; j++) {
			for(k = 0; k <= POL_DEG; k++) {
				c = poldi.coeff(x1, i).coeff(x2, j).coeff(x3, k);
				p_coeff[deg_conv(POL_DEG,i,j,k)] = ex_to<numeric>(c).to_double();
			}
		}
	}
	return;
}

void phi(INTERVAL &ret, INTERVAL *x)
{
	int i;
	INTERVAL temp_interval;
	
	pol_value(ret, POL_DEG, delta_coeff[0], x);
	ret = abs(ret);
	for(i = 1; i < 4; i++) {
		pol_value(temp_interval, POL_DEG, delta_coeff[i], x);
		temp_interval = abs(temp_interval);
		ret = boost::numeric::max(ret, temp_interval);
	}
	return;
}

void S6(INTERVAL &ret, INTERVAL *x)
{
	pol_value(ret, POL_DEG, S_coeff[0], x);
	return;
}

void S7(INTERVAL &ret, INTERVAL *x)
{
	pol_value(ret, POL_DEG, S_coeff[1], x);
	return;
}

void S8(INTERVAL &ret, INTERVAL *x)
{
	pol_value(ret, POL_DEG, S_coeff[2], x);
	return;
}

void S9(INTERVAL &ret, INTERVAL *x)
{
	pol_value(ret, POL_DEG, S_coeff[3], x);
	return;
}

void eq(INTERVAL &ret, INTERVAL *x)
{
	pol_value(ret, POL_DEG, pol_coeff, x);
	return;
}

int eq_check(INTERVAL *x)
{
	int deg, ret;
	INTERVAL xc[2], pol[POL_DEG_INC];
	
	xc[0] = median(x[0]);
	xc[1] = median(x[1]);
	
	deg = pol_subst(pol, POL_DEG, pol_coeff, xc);
	if(pol_sturm(deg, pol, x[2]) > 0) {
		ret = 1;
	} else {
		ret = 0;
	}

	return ret;
}

int main(void)
{
	int i, j, di;
	INTERVAL ret, bound;
	multiset<box> lb;
	box b;
	string buffer;
	numeric f_coeff[7];
	exmap ms;
	ex s[4];
	p_func p_ineq[4] = {S6,S7,S8,S9};
	p_func p_eq[1] = {eq};
	p_func_c p_eq_c[1] = {eq_check};

	ret.assign(0.0, 1.0);
	for(i = 0; i < POL_DEG_INC*POL_DEG_INC*POL_DEG_INC; i++) {
		pol_coeff[i] = 0.0;
		for(j = 0; j < 4; j++) {
			delta_coeff[j][i] = 0.0;
			S_coeff[j][i] = 0.0;
		}
	}
	
	for(i = 0; i < 3; i++) {
		b.interval[i].assign(-1.0, 1.0);
	}

	try {
	for(i=6; i >= 0; i--) {
		do {
		cout << "f" << i << ": " << flush;
		cin >> buffer;
		f_coeff[i] = buffer.c_str();
		} while(!f_coeff[i].is_integer());
	}
	} catch(exception &p) {
		cerr << p.what() << endl;
		return 1;
	}

	kummer k(f_coeff);
	s[0] = -(pow(k2,2)*pow(k3,2)*f_coeff[2]+(pow(k1,2)*pow(k3,2)
			-2*k1*pow(k2,2)*k3+pow(k2,4))*f_coeff[0]+k2*pow(k3,3)*f_coeff[3]
			+pow(k3,4)*f_coeff[4]+(-k1*k2*pow(k3,2)+pow(k2,3)*k3)*f_coeff[1]+k4*pow(k3,3));
	s[1] = -(pow(k1,2)*pow(k3,2)*f_coeff[2]+pow(k1,2)*pow(k2,2)*f_coeff[0]
			+pow(k3,4)*f_coeff[6]+pow(k1,2)*k2*k3*f_coeff[1]+k1*k4*pow(k3,2));
	s[2] = -(f_coeff[0] * pow(k1, 4) + (f_coeff[4] * pow(k3, 2)
			+ k4 * k3) * pow(k1, 2) + f_coeff[5] * pow(k3, 2) * k2 * k1
			+f_coeff[6] * pow(k3 * k2, 2));
	s[3] = -(pow(k1,4)*f_coeff[2]+(-pow(k1,2)*k2*k3+k1*pow(k2,3))*f_coeff[5]
			+(pow(k1,2)*pow(k3,2)-2*k1*pow(k2,2)*k3+pow(k2,4))*f_coeff[6]
			+pow(k1,3)*k2*f_coeff[3]+pow(k1,2)*pow(k2,2)*f_coeff[4]+pow(k1,3)*k4);

	do {
		cout << "i(1-4): " << flush;
		cin >> di;
	} while (di < 1 || di > 4);

	ex_mpz(pol_coeff, k.get_def_eq(), di);
	for(i = 0; i < 4; i++) {
		ex_mpz(delta_coeff[i], k.get_duplication(i+1), di);
		ex_mpz(S_coeff[i], s[i], di);
	}

	phi(b.value, b.interval);
	lb.insert(b);

	cout << scientific << setprecision(10);

	for(;;) {
		cout << "epsilon: ";
		cin >> buffer;
		optimize_epsilon = atof(buffer.c_str());
		if(optimize_epsilon <= 0.0) {
			return 0;
		}
		cout << "epsilon:" << optimize_epsilon << endl;
	
		if(minimum(ret, lb, phi, 4, p_ineq, 1, p_eq, p_eq_c) == 0) {
			cout << "No feasible points" << endl;
			return 0;
		}
	
		cout << ret << endl;
		
		if(ret.lower() > 0.0) {
			bound = log(ret) / (-3.0);
			cout << "upper bound: " << bound << endl;
		}

/*		
		cout << "number of intervals:" << lb.size() << endl;
		cout << "interval where the minimum attains:" << endl;
		for(i = 0; i < 3; i++) {
			cout << lb.begin()->interval[i] << endl;
		}
*/
	}


	return 0;
}
